/*
 * My small scheme
 *
 * $Id: s.c 1.107.1.23 Sat, 17 Apr 1999 00:42:42 +0200 crad $
 */
#include "s.h"
#include "vm2.h"
#include "heap.h"
#include "proc.h"
#include "stack.h"

/* #define OLD_INTERP */
/* #define SHOW_OPTIMIZE_PHASE */

/*-- arguments */
long scm_default_hs = 32;		/* default heap size in kilo cells */
long scm_default_ds = 32;		/* default data stack size in ptr to cells */
int scm_no_init=FALSE;			/* read default initfile */

char *scm_qs_lib;				/* QS_LIB environment variable */

/*-- cstack */
void *scm_cstack_start;			/* top of C stack (system stack) */

SOBJ scm_last_read_object;		/* last object read (used by err)*/


/*-- This table is used to describe various object type aspects.
 * execute - the code to be compiled when object of this type is used
 * 			 like a function call. ex (<object> ...)
 * name - the type name
 * mark - the code to mark object during GC
 * sweep- howto destroy object when not used
 * print- pretty print of object
 * write- debug print of object
 * creconizer - used by read to reconize this type
 * cparse - call by read after object has been reconized to parse it
 * wreconizer - used by read to reconize this type
 * wparse - call by read after object has been reconized to parse it
 * compare - how to compare of 2 objects of this type
 *
 * NOTE: this table must be keept in sync with SOBJ_TYPES enum in s.h
 */

SOBJ_TYPE_DESCR scm_type_hook[SOBJ_T_MAX] = {
  /* builtin types */
  /* execute,			name, 			mark, 			sweep,
	 print, 					write, 			
	 creconize, 				cparse,
	 wreconize, 				wparse,
	 compare */

  { SCM_OP_NO_CALL,		"void", 		},
  { SCM_OP_NO_CALL, 	"pair", 		},
  { SCM_OP_NO_CALL, 	"inum", 		},
  { SCM_OP_NO_CALL, 	"fnum", 		},
  { SCM_OP_NO_CALL, 	"bignum", 		},
  { SCM_OP_NO_CALL, 	"atom", 		scm_atom_mark,   scm_atom_sweep,
    scm_atom_print,		scm_atom_print  },
  { SCM_OP_NO_CALL, 	"keyword", 		},
  { SCM_OP_NO_CALL, 	"symbol", 		scm_symbol_mark, scm_symbol_sweep },
  { SCM_OP_NO_CALL, 	"lsymbol", 		},
  { SCM_OP_NO_CALL, 	"label", 		},

  { SCM_OP_NO_CALL, 	"module", 		scm_module_mark, scm_module_sweep,
	scm_module_print,	scm_module_write,
	NULL,				NULL,
	scm_module_wreconize,  scm_module_wparse
  },

  { SCM_OP_NO_CALL, 	"char", 		NULL,			NULL,				
	scm_char_print,				scm_char_write,
	scm_char_reconize,			scm_char_parse,
  },

  { SCM_OP_NO_CALL, 	"string", 		NULL,		   	scm_string_sweep,
	scm_string_print,			scm_string_write,
	scm_string_reconize,		scm_string_parse,
	NULL,						NULL,
	scm_string_compare
  },

  { SCM_OP_CALL_PRIM, 	"prim",			},
  { SCM_OP_CALL_CPRIM, 	"cprim", 		},

  { SCM_OP_NO_CALL, 	"syntax",		NULL,			NULL,
	scm_syntax_print },

  { SCM_OP_CALL_CODE, 	"code",			scm_code_mark, 	scm_code_sweep,
	scm_code_print 	},

  { SCM_OP_CALL_PROC, 	"proc",			scm_proc_mark, 	scm_proc_sweep,
	scm_proc_print 	},
  
  { SCM_OP_CALL_CLOSURE,"closure",		scm_clos_mark,	NULL,
	scm_clos_print	},
  
  { SCM_OP_NO_CALL, 	"env",			scm_env_mark,	scm_env_sweep,
	scm_env_print 	},
  
  { SCM_OP_CALL_MACRO, 	"macro",		scm_macro_mark,	NULL,
	NULL,			},
  
  { SCM_OP_NO_CALL, 	"port", 		NULL,			scm_port_sweep,
	scm_port_print 	},

  { SCM_OP_NO_CALL, 	"boolean", 		},
  { SCM_OP_NO_CALL, 	"unbound", 		},
  { SCM_OP_NO_CALL, 	"undefined", 	},

  { SCM_OP_NO_CALL, 	"eof",			NULL,			NULL,
	scm_eof_print,		scm_eof_write
  },
  
  { SCM_OP_NO_CALL, 	"continuation", scm_cont_mark,	scm_cont_sweep,
	NULL },
  
  { SCM_OP_NO_CALL, 	"array",		scm_array_mark,	scm_array_sweep,
	scm_vector_print,	scm_vector_write,
	scm_array_reconize, scm_array_parse,
	NULL,				NULL,
	NULL  },

  { SCM_OP_NO_CALL, 	"hash",			scm_hash_mark,	scm_hash_sweep,
	scm_hash_print, 	scm_hash_write, 	NULL  },

  { SCM_OP_NO_CALL, 	"pointer",		scm_pointer_mark,	scm_pointer_sweep,
	scm_pointer_print },
  
  { SCM_OP_CALL_EXTFUNC,"extfunc",		scm_extfunc_mark,	scm_extfunc_sweep,
	scm_extfunc_print,  scm_extfunc_write,	NULL },

  { SCM_OP_NO_CALL,		"extvar",		scm_extvar_mark,	scm_extvar_sweep,
	scm_extvar_print,   scm_extvar_write,	NULL },

  { SCM_OP_NO_CALL,		"var",			scm_var_mark,		scm_var_sweep,
	scm_var_print,   	scm_var_write,		NULL },

  { SCM_OP_CALL_VMFUNC,	"vmfunc",		scm_vmfunc_mark,	scm_vmfunc_sweep,
	NULL },

  { SCM_OP_NO_CALL,		"catch-context", scm_ccntxt_mark,	scm_ccntxt_sweep,
	NULL },
};

int scm_type_next_descr = SOBJ_T_USER;		/* next free descriptor */

/*-- predefined symbols */

SOBJ scm_true;					/* true */
SOBJ scm_false;					/* false */
SOBJ scm_unbound;				/* unbound */
SOBJ scm_undefined;				/* undefined  */

SOBJ scm_env,
  scm_sym_quote,
  scm_sym_qquote, scm_sym_unquote, scm_sym_unquote_splicing,
  scm_sym_set, scm_sym_if,
  scm_sym_lambda, scm_sym_begin, scm_sym_define,
  scm_sym_code, scm_sym_let, scm_sym_letstar, scm_sym_letrec,
  scm_sym_env,
  scm_sym_immediate,
  scm_sym_and, scm_sym_or,
  scm_sym_cond, scm_sym_else,
  scm_sym_while, 
  scm_sym_catch;

#ifdef OLD_COMPILER
void *scm_cons_addr, *scm_append_addr;
#endif

/*-- crash */
jmp_buf scm_errjmp;


/********************************
 * Stack
 ********************************/

void scm_stack_init(int size)
{
  scm_stack_size = size;
  scm_stack = scm_must_alloc(sizeof(SOBJ) * size);
  scm_stack_limit = scm_stack + size;
  sclear();
}

void sdump()
{
  SOBJ *p;

  scm_puts("stack: ");
#ifdef TOP_TO_BOTTOM
  p = scm_stack_limit;
  while(--p >= scm_sp) {
	scm_cdisplay(*p); scm_putc(' ');
  }
#else
  p = scm_sp;
  while(p < scm_stack_limit) {
	scm_cdisplay(*p++);  scm_putc(' ');
  }
#endif
  scm_putc('\n');
}

/********************************
 * Disassembler
 ********************************/

SOBJ scm_disasm(SOBJ obj)
{
  SCM_Code *c;
  SOBJ env, code;
  int i,j;
  
  switch(SCM_OBJTYPE(obj)) {
  case SOBJ_T_PROC:
	c = SCM_PROC_CODE(obj);
	scm_puts("; ");
	port_puts(scm_outp, "nargs="); port_putn(scm_outp, c->nargs);
	port_puts(scm_outp, " opt=");  port_putn(scm_outp, c->optargs);
	port_puts(scm_outp, " size="); port_putn(scm_outp, c->size);
	scm_puts("\n");
	scm_puts("(lambda ");
	env = c->envlist;
	scm_cdisplay(env);
	/*
	while(env) {
	  scm_cdisplay(SCM_CAR(env));  scm_cdisplay(SCM_CAAR(env));  
	  env = SCM_CDR(env);
	  if (env) scm_putc(' ');
	}
	*/
	scm_puts("\n");
	scm_disassemble(c->code, c->size);
	break;
  case SOBJ_T_CODE:
	scm_puts("(code\n");
	scm_puts("; ");
	port_puts(scm_outp, "size=");  port_putn(scm_outp, SCM_CODE_SIZE(obj));
	scm_puts("\n");
	scm_disassemble(SCM_CODE_CODE(obj), SCM_CODE_SIZE(obj));
	break;
  case SOBJ_T_CLOSURE:
	code = SCM_CLOSURE_CODE(obj);
	env  = SCM_CLOSURE_ENV(obj);
	scm_puts("; closure environment:\n");
	j = 0;
	while(env) {
	  scm_puts("; level #"); scm_putn(j); scm_puts(": ");
	  if (SCM_ENV_FRAME(env) == NULL) {
		scm_puts("**no frame**");
	  } else {
		for (i = 0; i < SCM_INUM(SCM_ENV_FRAME(env)->nslots); i++) {
		  scm_cdisplay(SCM_ENV_FRAME(env)->binding[i]);  scm_puts(" ");
		}
	  }
	  scm_puts("\n");
	  env = SCM_ENV_NEXT(env);
	  j++;
	}
	scm_disasm(code);
	return(scm_undefined);
  default:
	err("disasm: bad closure, code or proc", obj);
  }
  scm_puts(")\n");
  return(scm_undefined);
}


/********************************
 * Reader
 ********************************/

static char scm_tokbuf[1024];
static SOBJ scm_tokval;			/* for user defined tokens */

enum SCM_TOKENS {
  SCM_TOK_EOF = 0,
  SCM_TOK_LPAR,
  SCM_TOK_RPAR,
  SCM_TOK_DOT,
  SCM_TOK_NUM,
  SCM_TOK_SYM,
  SCM_TOK_SHARP,
  SCM_TOK_BOOLEAN,
  SCM_TOK_VECTOR,
  SCM_TOK_QUOTE,
  SCM_TOK_QQUOTE,
  SCM_TOK_UNQUOTE,
  SCM_TOK_UNQUOTE_SPLICING,
  SCM_TOK_DQUOTE,
  SCM_TOK_EVAL,
  SCM_TOK_UDEF
};


/* jump over blanks: return the first non blank char */
int scm_skip_blanks(PORT *port)
{
  int c;
  while((c = port_getc(port)) != EOF && isspace(c))
	;
  return(c);
}

/* get next word to tokbuf */
int scm_get_word(PORT *port, int c)
{
  char *p;
  p = scm_tokbuf;
  *p++ = c;
  while( (c = port_peekc(port)) != EOF ) {
	if (isspace(c) || c == '(' || c == ')') 	break;
	*p++ = port_getc(port);
  }
  *p = 0;
  return(c);
}

int scm_read_token(PORT *port)
{
  int c, next, i;

  /* ignore comments: ';' and '#!' comments a line */
  while( (c = scm_skip_blanks(port)) == ';' ||
		 (c == '#' && port_peekc(port) == '!')) {
	while((c = port_getc(port)) != '\n' && c != EOF)
	  ;
  }

  if (c == EOF) {	return(SCM_TOK_EOF);  }
  if (c == '(') {	return(SCM_TOK_LPAR); }
  if (c == ')') {	return(SCM_TOK_RPAR); }
  if (c == '\'') {	return(SCM_TOK_QUOTE); }
  if (c == '`') {	return(SCM_TOK_QQUOTE);	 }
  if (c == ',') {
	next = port_peekc(port);
	if (next != '@') 	return(SCM_TOK_UNQUOTE);
	port_getc(port);
	return(SCM_TOK_UNQUOTE_SPLICING);
  }
  if (c == '.') {
	next = port_peekc(port);
	if (isspace(next))	return(SCM_TOK_DOT);
  }

  if (c == '#') {
	next = port_peekc(port);
	if (strchr("bodxie", next)) { /* number marker */
	  scm_get_word(port, c);
	  return(SCM_TOK_NUM);		/* signal number */
	}
	if (next == 't' || next == 'f') {
	  scm_tokbuf[0] = next;
	  scm_tokbuf[1] = 0;
	  port_getc(port);
	  return(SCM_TOK_BOOLEAN);
	}
	/* jump over the next char (just peeked) */	
	if (next == '.') {
	  port_getc(port);
	  return(SCM_TOK_EVAL);
	}
  }
  for (i = 0; i < scm_type_next_descr; i++) {
	if (scm_type_hook[i].creconize != NULL &&
		(*scm_type_hook[i].creconize)(port, c)) {
	  scm_tokval = (*scm_type_hook[i].cparse)(port, c);
	  return(SCM_TOK_UDEF);
	}
  }

  scm_get_word(port, c);

  if (isdigit((int)scm_tokbuf[0])) {		return(SCM_TOK_NUM); }
  if (scm_tokbuf[0] == '-' &&
	  isdigit((int)scm_tokbuf[1])) {		return(SCM_TOK_NUM); }

  if (scm_tokbuf[0] == '.' &&
	  isdigit((int)scm_tokbuf[1])) {		return(SCM_TOK_NUM); }

  if (scm_tokbuf[0] == ':') {
	scm_tokval = scm_keyword_add(scm_tokbuf);
	return(SCM_TOK_UDEF);
  }

  for (i = 0; i < scm_type_next_descr; i++) {
	if (scm_type_hook[i].wreconize != NULL &&
		(*scm_type_hook[i].wreconize)(port, scm_tokbuf)) {
	  scm_tokval = (*scm_type_hook[i].wparse)(port, scm_tokbuf);
	  return(SCM_TOK_UDEF);
	}
  }
  return(SCM_TOK_SYM);
}

SOBJ scm_read_list(PORT *port)
{
  int tok;
  SOBJ list, last, obj;

  list = last = obj = NULL;

  while( (tok = scm_read_token(port)) != SCM_TOK_RPAR) {
	switch(tok) {
	case SCM_TOK_EOF:
	  fprintf(stderr, "scm_read_list: EOF while reading list\n");
	  return(scm_eof);

	case SCM_TOK_NUM:
	  obj = scm_cstr2number(scm_tokbuf, 10);
	  if (obj != scm_false)
		break;
	  /* FALL THROUGH */

	case SCM_TOK_SYM:	obj = scm_mkatom(scm_tokbuf);				break;
	case SCM_TOK_LPAR:	obj = scm_read_list(port);					break;
	case SCM_TOK_DOT:
	  if (SCM_NNULLP(last)) {
		SCM_CDR(last) = scm_read_port(port);
		if (scm_read_token(port) != SCM_TOK_RPAR) {
		  fprintf(stderr, "scm_read_list: illegal '.' expression\n");
		  return(NULL);
		}
		return(list);
	  }
	  err("scm_read_list: unexpected '.'", NULL);

	case SCM_TOK_SHARP:			/* special form */
	  break;

	case SCM_TOK_BOOLEAN:
	  obj = (scm_tokbuf[0] == 'f') ? scm_false : scm_true;
	  break;

	case SCM_TOK_UDEF:
	  obj = scm_tokval;
	  break;

	case SCM_TOK_QUOTE:	obj = SCM_LIST2(scm_sym_quote,  scm_read_port(port)); break;
	case SCM_TOK_QQUOTE:obj = SCM_LIST2(scm_sym_qquote, scm_read_port(port)); break;
	case SCM_TOK_UNQUOTE:
	  obj = SCM_LIST2(scm_sym_unquote, scm_read_port(port));
	  break;

	case SCM_TOK_UNQUOTE_SPLICING:
	  obj = SCM_LIST2(scm_sym_unquote_splicing, scm_read_port(port));
	  break;
	  
	case SCM_TOK_EVAL:
	  obj = scm_eval(scm_read_port(port), NULL);
	  break;

	default:
	  
	  fprintf(stderr, "scm_read_list: unreconized token %d\n", tok);
	  exit(1);

	}
	if (SCM_NNULLP(last)) {
	  SCM_CDR(last) = scm_cons(obj, NULL);
	  last = SCM_CDR(last);
	} else {
	  list = last = scm_cons(obj, NULL);
	}
  }
  return(list);
}

SOBJ scm_read_port(PORT *port)
{
  int tok;

  if (port == NULL) {  port = scm_inp; }

  tok = scm_read_token(port);
  switch(tok) {
  case SCM_TOK_EOF:		return(scm_eof);
  case SCM_TOK_LPAR:	return(scm_read_list(port));
	
  case SCM_TOK_NUM:
	{
	  SOBJ obj;
	  obj = scm_cstr2number(scm_tokbuf, 10);
	  if (obj != scm_false)	return(obj);
	}
	/* FALL THROUGH */
	
  case SCM_TOK_SYM:		return(scm_mkatom(scm_tokbuf));
  case SCM_TOK_BOOLEAN:	return(scm_tokbuf[0] == 'f' ? scm_false : scm_true);
  case SCM_TOK_RPAR:	err("scm_read: unexpected ')'", NULL);
  case SCM_TOK_DOT:		err("scm_read: unexpected '.'", NULL);
  case SCM_TOK_UDEF:	return(scm_tokval);

  case SCM_TOK_QUOTE:	return(SCM_LIST2(scm_sym_quote,  scm_read_port(port)));
  case SCM_TOK_QQUOTE:	return(SCM_LIST2(scm_sym_qquote, scm_read_port(port)));
  case SCM_TOK_UNQUOTE:
	return(SCM_LIST2(scm_sym_unquote, scm_read_port(port)));

  case SCM_TOK_UNQUOTE_SPLICING:
	return(SCM_LIST2(scm_sym_unquote_splicing, scm_read_port(port)));

  case SCM_TOK_EVAL:	return(scm_eval(scm_read_port(port), NULL));
	
  default:
	err("scm_read: unreconized token", NULL);
  }
  return(NULL);
}

/********************************
 * compiler
 ********************************/

#ifdef OLD_COMPILER
static void scm_syntax_lset(SOBJ lvar, int depth)
{
  SOBJ ofs = SCM_MKINUM(SCM_LSYM_OFS(lvar));
  switch(depth) {
  case 0:	spush(ofs);	spush_opc(SCM_OP_SETL0);  break;
  default:
	spush( SCM_MKINUM((depth << 16) | SCM_LSYM_OFS(lvar)));
	spush_opc(SCM_OP_SETL);
  }
}

static void scm_syntax_lget(SOBJ lvar, int depth)
{
  SOBJ ofs = SCM_MKINUM(SCM_LSYM_OFS(lvar));
  switch(depth) {
  case 0:	spush(ofs); spush_opc(SCM_OP_PUSHL0); break;
  case 1:	spush(ofs); spush_opc(SCM_OP_PUSHL1); break;
  case 2:	spush(ofs); spush_opc(SCM_OP_PUSHL2); break;
  case 3:	spush(ofs); spush_opc(SCM_OP_PUSHL3); break;
  default:
	spush( SCM_MKINUM((depth << 16) | SCM_LSYM_OFS(lvar)));
	spush_opc(SCM_OP_PUSHL);
  }
}

SOBJ scm_syntax_object(SOBJ obj, SOBJ env);
SOBJ scm_syntax_pair(SOBJ obj, SOBJ env);

SOBJ scm_syntax_code(SOBJ list, SOBJ env)
{
  SOBJ obj;
  void *addr;
  
  /*-- recurse to end of list */
  if (list == NULL) return(NULL);
  scm_syntax_code(SCM_CDR(list), env);
  
  /*  scm_puts("Compile code:\n");
	  printf("List length=%d\n", scm_list_length(list)); */

  obj = SCM_CAR(list);
  switch(SCM_OBJTYPE(obj)) {
  case SOBJ_T_SYMBOL:
	scm_puts("symbol: "); scm_cprint(obj);
	addr = scm_search_opcode_address(SCM_ATOM_NAME(SCM_SYM_NAME(obj)));
	if (addr != NULL) {
	  spush(addr);
	  break;
	}
	if (env) {
	  int depth;
	  SOBJ lvar = scm_env_search(env, SCM_SYM_NAME(obj), &depth);
	  if (lvar) {
		scm_syntax_lget(lvar, depth);
	  }
	  break;
	}
	spush(obj);
	break;

  case SOBJ_T_INUM:
	scm_puts("number: "); scm_cprint(obj);
	spush(obj);
	break;
	
  default:
	scm_syntax_object(obj, env);
  }
  return(NULL);
}

/*-- compiler for quote */
SOBJ scm_syntax_quote(SOBJ args, SOBJ env)
{
  spush(SCM_CAR(args));  spush_opc(SCM_OP_PUSH);
  return(NULL);
}

static SOBJ scm_syntax_qq1(SOBJ args, SOBJ env)
{
  SOBJ obj, item, list, last;

  /* scm_puts("; qq:"); scm_cprint(args);  */

  if (args == NULL)			return(NULL);
  if (!SCM_PAIRP(args))	err("scm_syntax_qq1: expected pair", args);
  
  if (SCM_CAR(args) == scm_sym_unquote) {
	scm_syntax_object(SCM_CADR(args), env);
	return(NULL);
  }
  last = NULL;
  list = NULL;
  for (obj = args; obj; obj = SCM_CDR(obj)) {
	item = SCM_CAR(obj);
	if (item && SCM_PAIRP(item)) {
	  if (list) {
		spush(scm_append_addr);  spush(list);  spush_opc(SCM_OP_PUSH);
		list = NULL;  last = NULL;
	  }
	  if (SCM_CAR(item) == scm_sym_unquote) {
		spush(scm_cons_addr);
		scm_syntax_object(SCM_CADR(item), env);
		continue;
	  }
	  if (SCM_CAR(item) == scm_sym_unquote_splicing) {
		spush(scm_append_addr);
		scm_syntax_object(SCM_CADR(item), env);
		continue;
	  }
	  if (SCM_CAR(item) != scm_sym_qquote) {
		spush(scm_cons_addr);
		scm_syntax_qq1(item, env);
		/*		spush(NULL);  spush_opc(SCM_OP_PUSH); */
		continue;
	  }
	}
	if (list == NULL) {
	  list = scm_cons(item, NULL);
	  last = list;
	} else {
	  SCM_CDR(last) = scm_cons(item, NULL);
	  last = SCM_CDR(last);
	}
  }
  spush(list);  spush_opc(SCM_OP_PUSH);
  return(NULL);
}

SOBJ scm_syntax_qquote(SOBJ args, SOBJ env)
{
  scm_syntax_qq1(SCM_CAR(args), env);
  /* sdump(); */
  return(NULL);
}

static int scm_syntax_args(SOBJ arglist, SOBJ env)
{
  int n = 0;

  while(SCM_NNULLP(arglist)) {
	scm_syntax_object(SCM_CAR(arglist), env);
	arglist = SCM_CDR(arglist);
	n++;
  }
  return(n);
}

/*-- compiler for set */
SOBJ scm_syntax_set2(SOBJ sym, SOBJ expr, SOBJ env)
{
  int depth;
  SOBJ var;
  
  if (!SCM_ATOMP(sym))
	err("set!: strange symbol", sym);

  if (env != NULL) {			/* we have an environment chain */
	var = scm_env_search(env, sym, &depth);
	if (var != NULL) {			/* found local variable */
	  scm_syntax_lset(var, depth);
	  scm_syntax_args(expr, env);
	  return(NULL);
	}
  }
  /*-- must be defined in the global env */
  var = scm_module_symbol_search(scm_current_module, sym);

  if (var == NULL) err("set!: undefined symbol", sym);
  
  spush_opc(SCM_OP_STORE);
  spush(var);
  spush_opc(SCM_OP_PUSH);
  scm_syntax_args(expr, env);
  return(NULL);

}

SOBJ scm_syntax_set(SOBJ args, SOBJ env)
{
  return(scm_syntax_set2(SCM_CAR(args), SCM_CDR(args), env));
}

/*-- compiler for if */
static SOBJ scm_syntax_if(SOBJ args, SOBJ env)
{
  int l = scm_list_length(args);
  SOBJ tst, tpart, fpart;
  SOBJ *b1, *b2;
  long ofs;
  
  if (l != 2 && l != 3) {
	fprintf(stderr, "syntax: (if <test> <truepart> <falsepart>)\n");
	return(NULL);
  }

  tst = SCM_CAR(args);
  tpart = SCM_CAR(SCM_CDR(args));
  fpart = (l == 3) ? SCM_CAR(SCM_CDR(SCM_CDR(args))) : (SOBJ)NULL;

  b2 = scm_sp;
  if (SCM_NNULLP(fpart)) {
	scm_syntax_object(fpart, env);
  } else {
	spush(NULL);				/* default fpart : push NIL */
	spush_opc(SCM_OP_PUSH);
  }
  b1 = scm_sp;
  ofs = b2 - scm_sp;	spush( SCM_MKINUM(ofs + 1) );
  spush_opc(SCM_OP_BRA);
  scm_syntax_object(tpart, env);
  ofs = b1 - scm_sp;	spush( SCM_MKINUM(ofs + 1) );
  spush_opc(SCM_OP_BRF);
  
  scm_syntax_object(tst, env);
	
  return(NULL);
}

/*-- begin:
 *
 * (begin <expr> ... )
 *
 * Compile a list of expressions. Generate a link before and an unlink after
 * to warranty a unique result. The result is the value of the last expr.
 */
SOBJ scm_syntax_begin(SOBJ exprlist, SOBJ env)
{
 
  /*  scm_puts("; scm_syntax_begin: exprlist=");  scm_cprint(exprlist); */
  if (exprlist == NULL)	return(NULL);
  
  scm_syntax_begin(SCM_CDR(exprlist), env);
  if (SCM_CDR(exprlist) != NULL) {	spush_opc(SCM_OP_DROP);  }
  scm_syntax_object(SCM_CAR(exprlist), env);
  return(NULL);
}

/* -- code optimizer
 *
 *
 */
static void optimize_code(SOBJ *start, SOBJ *limit)
{
  SOBJ *p;
  int changes, passes;

  changes = 1; passes = 0;
  while(changes) {
	passes++;
	changes = 0;
	for (p = start; p < limit; p++) {
	  if (p[0] == opcode(SCM_OP_BRA) &&
		  p[SCM_INUM(p[1]) + 1] == opcode(SCM_OP_RETURN)) {
		
#ifdef SHOW_OPTIMIZE_PHASE
		printf("; optimize: bra to return -> return\n");
#endif
		p[0] = opcode(SCM_OP_RETURN);
		p[1] = opcode(SCM_OP_NOP); /* cleanup */
		changes = 1;
		break;
	  }

	  if (p[0] == opcode(SCM_OP_CALL) && p < (limit - 2) &&
		  p[1] == opcode(SCM_OP_ENDLET) &&
		  p[2] == opcode(SCM_OP_RETURN)) {

#ifdef SHOW_OPTIMIZE_PHASE
		printf("; optimize: call endlet return->endlet-jump endlet return\n");
#endif
		p[0] = opcode(SCM_OP_ENDLET_JUMP);
		changes = 1;
		break;
	  }

	  if (p[0] == opcode(SCM_OP_CALL) && p < (limit - 1) &&
		  p[1] == opcode(SCM_OP_RETURN)) {

#ifdef SHOW_OPTIMIZE_PHASE
		printf("; optimize: call -> jump\n");
#endif
		p[0] = opcode(SCM_OP_JUMP);
		changes = 1;
		break;
	  }
	}
  }
#ifdef SHOW_OPTIMIZE_PHASE
  printf("; optimize: %d passes\n", passes);
#endif
}


/*-- lambda:
 *
 * (lambda <formal> <body>)
 *
 * The strategy to compile lambda is:
 * - remember current sp. This will be the last instruction of the lambda
 * - compile the lambda
 * - make a new procedure object
 * - copy the procedure code to a newly allocated memory array.
 * - restore sp
 * - push the number of args and the addr of the procedure object
 * - that's all
 *
 */
static SOBJ scm_syntax_lambda2(SOBJ formal, SOBJ body, SOBJ env)
{
  SOBJ proc, *oldsp, newenv;
  SCM_Code *code;
  int fnr;						/* formal number */
  int optarg;					/* flag : t when opt args */

  /* debug
  scm_puts("; scm_syntax_lambda: formal="); scm_cdisplay(formal);
  scm_puts(" body="); scm_cprint(body);
  if (*scm_sp != opcode(SCM_OP_CALL)) {
	scm_puts("; WARNING: lambda not called => mkclosure\n");
  }
  */

  oldsp  = scm_sp;				/* the stack before compiling lambda */

  proc = scm_newcell(SOBJ_T_PROC);

  /*-- build the new environment */
  newenv = scm_env_add_level(env);
  fnr = 0;
  while(formal && SCM_PAIRP(formal)) {
	newenv = scm_env_add(newenv, scm_mklsymbol(SCM_CAR(formal), fnr++));
	formal = SCM_CDR(formal);
  }

  optarg = FALSE;
  if (formal && SCM_ATOMP(formal)) {
	newenv = scm_env_add(newenv, scm_mklsymbol(formal, fnr++));
	optarg = TRUE;
  }
  /*
	scm_puts("; ct-env = "); 	scm_cprint(newenv);
	scm_puts("; nargs="); 	scm_putn(fnr); 
	scm_puts(" optarg=");		scm_putn(optarg);
  */

  /*-- code generation starts here */
  spush_opc(SCM_OP_RETURN);		/* oldsp[-1] */
  scm_syntax_begin(body, newenv);
  optimize_code(scm_sp, oldsp);

  /*-- create the code structure and fill it*/

  /*
	scm_puts(" code size="); scm_putn(oldsp - scm_sp); scm_putc('\n');
    scm_puts("; ");  sdump();
  */
  
  code = scm_must_alloc(  offsetof(SCM_Code, code[oldsp - scm_sp]) );
  code->envlist = newenv;
  code->nargs   = fnr;
  code->optargs = optarg;
  code->nlocals = 0;
  code->size = oldsp - scm_sp;
  memcpy(code->code, scm_sp, code->size * sizeof(SOBJ*));

  SCM_PROC_CODE(proc) = code;
  SCM_PROC_ENV(proc) = NULL;

  scm_sp = oldsp;
  spush_opc(SCM_OP_MKCLOSURE);  spush(proc);  spush_opc(SCM_OP_PUSH);
  return(NULL);
}

SOBJ scm_syntax_lambda(SOBJ arglist, SOBJ env)
{
  return(scm_syntax_lambda2(SCM_CAR(arglist), SCM_CDR(arglist), env));
}

/*-- compiler for define */
SOBJ scm_syntax_define(SOBJ alist, SOBJ env)
{
  SOBJ var, body;

  var = SCM_CAR(alist);
  body = SCM_CDR(alist);

  if (SCM_ATOMP(var)) {			/* (define <var> <value> ) */
	/* force creation of the symbol */
	scm_module_symbol_lookup(scm_current_module, var);
	return(scm_syntax_set2(var, body, env));
  }

  if (SCM_PAIRP(var)) {			/* (define (var formals) body) */
	/* force creation of the symbol */
	scm_module_symbol_lookup(scm_current_module, SCM_CAR(var));
	scm_syntax_set2(SCM_CAR(var), NULL, env);
	scm_syntax_lambda2(SCM_CDR(var), body, env);
	return(NULL);
  }
  err("define: bad syntax. Atom or pair wanted, got", var);
  return(NULL);
}

/* The let compiler
 *
 * expr:
 *
 *  (let ( [ <binding> ] ) <body> )
 *
 *	<binding> = ( <var> [ <init> ] )
 *
 * Compiled like that:
 *
 * 	spush_opc(SCM_OP_UNLINKENV)
 *	<body expressions>
 * 	spush(proc_node);
 *  spush_opc(SCM_OP_LINKENV)
 *  <init expressions>
 */
static SOBJ scm_syntax_let1(SOBJ arglist, int star, SOBJ env)
{
  SOBJ binding, body;
  SOBJ obj, pair, newenv;
  int fnr;

  binding = SCM_CAR(arglist);
  body = SCM_CDR(arglist);
  
  /* debug
	 scm_puts("; let: binding="); scm_cdisplay(binding);
	 scm_puts(" body="); scm_cprint(body);
  */
  if (binding == NULL) 	err("(let <binding> ...): binding missing", arglist);
  if (body == NULL)		err("(let <binding> <body>): body missing", arglist);
  
  obj = binding;
  newenv = scm_env_add_level(env);
  fnr = 0;
  while(obj) {
	pair = SCM_CAR(obj);
	if (!SCM_PAIRP(pair)) 			 err("let: not a pair", obj);
	if (!SCM_ATOMP(SCM_CAR(pair)))
	  err("let: invalid variable name", SCM_CAR(pair));

	/* debug
	   if (!SCM_SYMBOLP(SCM_CAR(pair)))err("let: not a symbol", SCM_CAR(pair));
	   scm_puts("; let binding ");  scm_cprint(pair);
	*/
	newenv = scm_env_add(newenv, scm_mklsymbol(SCM_CAR(pair), fnr++));
	obj = SCM_CDR(obj);
  }
  /* scm_puts("; let env=");  scm_cprint(newenv); */
  
  /*-- codegen : body */

  spush_opc(SCM_OP_ENDLET);
  scm_syntax_begin(body, newenv);

  if (!star) {
	spush_opc(SCM_OP_DOLET);

	/*-- codegen : init */
	for (obj = binding; obj; obj = SCM_CDR(obj)) {
	  pair = SCM_CAR(obj);
	  if (!SCM_PAIRP(pair)) err("need a pair", obj);
	  if (SCM_CDR(pair)) {
		scm_syntax_object(SCM_CADR(pair), env);
	  } else {
		spush(scm_unbound);	spush_opc(SCM_OP_PUSH);
	  }
	}
	spush_opc(SCM_OP_MARK);
  } else {
	SOBJ list;
	list = NULL;
	for (obj = binding; obj; obj = SCM_CDR(obj)) 
	  list = scm_cons(SCM_CAR(obj), list);
	
	for (obj = list; obj; obj = SCM_CDR(obj)) {
	  pair = SCM_CAR(obj);
	  if (!SCM_PAIRP(pair)) err("let* : binding must be pairs...", pair);
	  spush_opc(SCM_OP_DROP);
	  scm_syntax_set2(SCM_CAR(pair), SCM_CDR(pair), newenv);
	}
	spush(SCM_MKINUM(fnr));  spush_opc(SCM_OP_DOLETSTAR);
  }
  return(NULL);
}

SOBJ scm_syntax_let(SOBJ obj, SOBJ env)
{
  return(scm_syntax_let1(obj, FALSE, env));
}

static SOBJ scm_syntax_lets(SOBJ obj, SOBJ env)
{
  return(scm_syntax_let1(obj, TRUE, env));
}

static SOBJ scm_syntax_the_env(SOBJ obj, SOBJ env)
{
  spush(env);  spush_opc(SCM_OP_PUSH);
  return(obj);
}

static SOBJ scm_syntax_immediate(SOBJ obj, SOBJ env)
{
  SOBJ value;
  scm_puts("; immediate\n");
  scm_puts("; obj=");  scm_cprint(obj);
  scm_puts("; env=");  scm_cprint(env);
  value = scm_compile(SCM_CAR(obj), env);
  scm_puts("; value="); scm_cprint(value); scm_disasm(value);
  spush_opc(SCM_OP_CALL);
  spush(value);
  spush_opc(SCM_OP_PUSH);
  spush_opc(SCM_OP_MARK);
  return(obj);
  
  value = scm_eval(SCM_CAR(obj), env);
  scm_puts("; eval returns: ");  scm_cprint(value);
  return(scm_syntax_object(value, env));
}

/*-- syntax and */

static SOBJ scm_syntax_and2(SOBJ obj, SOBJ env, SOBJ *the_jump)
{
  if (obj == NULL) return(NULL);

  scm_syntax_and2(SCM_CDR(obj), env, the_jump);
  if (SCM_CDR(obj)) {
	spush(SCM_MKINUM( (the_jump - scm_sp) ));
	spush_opc(SCM_OP_BR_AND);
  }
  scm_syntax_object(SCM_CAR(obj), env);
  return(NULL);
}

static SOBJ scm_syntax_and(SOBJ obj, SOBJ env)
{
  if (obj == NULL) {
	spush(scm_true);  spush_opc(SCM_OP_PUSH);  return(NULL);
  }
  return(scm_syntax_and2(obj, env, scm_sp));
}


/*-- syntax or */

SOBJ scm_syntax_or2(SOBJ obj, SOBJ env, SOBJ *the_jump)
{
  if (obj == NULL) return(NULL);

  scm_syntax_or2(SCM_CDR(obj), env, the_jump);
  if (SCM_CDR(obj)) {
	spush(SCM_MKINUM( (the_jump - scm_sp) ));
	spush_opc(SCM_OP_BR_OR);
  }
  scm_syntax_object(SCM_CAR(obj), env);
  return(NULL);
}

SOBJ scm_syntax_or(SOBJ obj, SOBJ env)
{
  if (obj == NULL) {
	spush(scm_false);  spush_opc(SCM_OP_PUSH);  return(NULL);
  }
  return(scm_syntax_or2(obj, env, scm_sp));
}


/*-- syntax cond */

/* syntax = (cond <clause1> <clause2>...[<else>])
 * clause = (<test> <expr1>...)
 * else   = (else <expr1>...)
 */	
   
SOBJ scm_syntax_cond2(SOBJ obj, SOBJ env, SOBJ *the_jump)
{
  SOBJ *spsave;
  SOBJ clause;

  if (obj == NULL) return(NULL);
  
  scm_syntax_cond2(SCM_CDR(obj), env, the_jump);
  
  if (SCM_CAR(obj) == scm_sym_else) {
	if (SCM_CDR(obj) != NULL)
	  err("else clause not in last position", obj);
  } else {
	spsave = scm_sp;
	if (SCM_CDR(obj) != NULL) {
	  spush(SCM_MKINUM(the_jump - scm_sp));
	  spush_opc(SCM_OP_BRA);
	}
	clause = SCM_CAR(obj);
	if (!SCM_PAIRP(clause)) err("malformed clause", clause);

	scm_syntax_begin(SCM_CDR(clause), env);
	if (SCM_CAR(clause) != scm_sym_else) {
	  spush(SCM_MKINUM(spsave - scm_sp));
	  spush_opc(SCM_OP_BRF);
	  scm_syntax_object(SCM_CAR(clause), env);
	}
  }
  return(NULL);
}

SOBJ scm_syntax_cond(SOBJ obj, SOBJ env)
{
  return(scm_syntax_cond2(obj, env, scm_sp));
}

/*-- catch:
 *
 * (catch TAG HANDLER EXPR)
 */

SOBJ scm_syntax_catch(SOBJ body, SOBJ env)
{
  SOBJ tag, handler;
  SOBJ *spsave;

  if (!SCM_PAIRP(body))	err("catch: bad tag in", body);
  tag = SCM_CAR(body);  body = SCM_CDR(body);

  if (!SCM_PAIRP(body))	err("catch: bad handler in", body);
  handler = SCM_CAR(body);  body = SCM_CDR(body);

  scm_puts("; catch: body=");  	scm_cdisplay(body);
  scm_puts(" env=");  			scm_cprint(env);

  spush_opc(SCM_OP_UNCATCH);  spsave = scm_sp;
  spush_opc(SCM_OP_RETURN);
  scm_syntax_begin(body, env);

  spush(SCM_MKINUM(spsave - scm_sp));
  spush_opc(SCM_OP_CATCH);
  scm_syntax_object(tag, env);
  scm_syntax_object(handler, env);

  return(body);
}

/*-- syntax: case */
/*
 * (case <key_expr> [ ( <datum> <expr> ...) ]+ [ (else <expr>...) ])
 */


/*-- syntax: macro */

SOBJ scm_syntax_macro(SOBJ symv, SOBJ form, SOBJ env)
{
  SOBJ code, result;
  SOBJ *oldsp;
  
  code = SCM_MACRO_CODE(symv);
  
  oldsp = scm_sp;
  spush_opc(SCM_OP_END);
  spush_opc(SCM_OP_CALL);
  spush(code);  spush_opc(SCM_OP_PUSH);
  spush(form);	spush_opc(SCM_OP_PUSH);
  spush_opc(SCM_OP_MARK);

  spush(NULL);
  scm_start_engine();
  result = spop();
  /* scm_puts("; macro result=");  scm_cprint(result); */
  scm_sp = oldsp;
  return(result);
}

/* The pair compiler
 *
 * Expression like:
 *
 * 	( s1 s2 .. )
 *
 * Are compiled with following sequence:
 *
 * 	spush_opc(SCM_OP_CALL);
 * 	spush(s1)
 * 	spush_opc(SCM_OP_PUSH);
 * 	...
 * 	spush_opc(SCM_OP_LINK);
 *
 * Giving the following code:
 *
 *     link push sn ... push s2 push s1 call unlink
 *
 * Note: eval if it's possible to merge call and unlink in a single opcode
 */
SOBJ scm_syntax_pair(SOBJ form, SOBJ env)
{
  SOBJ func, arglist, symv;
  int  nargs, needlink;

  func = SCM_CAR(form);		/* start obj */
  arglist  = SCM_CDR(form);		/* rest of list */

  needlink = TRUE;

  switch(SCM_OBJTYPE(func)) {
  case SOBJ_T_PAIR:
	spush_opc(SCM_OP_CALL);
	scm_syntax_pair(func, env);
	break;
  case SOBJ_T_INUM:
  case SOBJ_T_BNUM:
  case SOBJ_T_FNUM:
	fprintf(stderr, "illegal expr: (number ...)\n");
	return(NULL);
  case SOBJ_T_ATOM:				/* see comment in scm_syntax_object */
	{
	  SOBJ v;
	  int depth;
	  if (env) {
		if ((v = scm_env_search(env, func, &depth)) != NULL) {
		  spush_opc(SCM_OP_CALL);
		  scm_syntax_lget(v, depth);
		  break;
		}
	  }
	  v = scm_module_find_symbol(scm_current_module, func, TRUE);
	  func = v;
	  /* fall through */
	}
  case SOBJ_T_SYMBOL:			/* symbol */
	if (env) {
	  int depth;
	  SOBJ lvar = scm_env_search(env, SCM_SYM_NAME(func), &depth);
	  if (lvar) {
		spush_opc(SCM_OP_CALL);
		scm_syntax_lget(lvar, depth);
		break;
	  }
	} 
	symv = SCM_SYM_VALUE(func);
	if (symv == NULL) {  symv = scm_unbound; }
	switch(SCM_OBJTYPE(symv)) {
	case SOBJ_T_CPRIM:
	  if (SCM_CPRIM_NARGS(symv) >= 0) {	/* fixed number of args */
		if (SCM_CPRIM_NARGS(symv) > 5) {
		  printf("Unsupported primitive with %d args\n",
				 SCM_CPRIM_NARGS(symv));
		  longjmp(scm_errjmp, 1);
		}
		spush_opc(SCM_OP_CALLC0 + SCM_CPRIM_NARGS(symv));
		spush(symv);
		spush_opc(SCM_OP_PUSH);
		needlink = FALSE;
	  } else {					/* will take args directly from stack */
		spush_opc(SCM_OP_CALLS);
		spush(symv);
		spush_opc(SCM_OP_PUSH);
	  }
	  break;
	case SOBJ_T_PRIM:			/* primitive */
	  spush(SCM_PRIM(symv)->address);
	  needlink = (SCM_PRIM(symv)->nargs < 0);
	  break;

	case SOBJ_T_SYNTAX:			/* syntax object. immediate execution */
	  {
		return((*SCM_SYNTAX_FUNC(symv))(arglist, env));
	  }
	case SOBJ_T_MACRO:
	  {
		return(scm_syntax_object(scm_syntax_macro(symv, form, env), env));
	  }

	default:
	  spush_opc(SCM_OP_CALL);
	  spush(func);
	  spush_opc(SCM_OP_PUSHV);
	}
	break;

  case SOBJ_T_CPRIM:
	if (SCM_CPRIM_NARGS(func) >= 0) {	/* fixed number of args */
	  if (SCM_CPRIM_NARGS(func) > 5) {
		printf("Unsupported primitive with %d args\n",
			   SCM_CPRIM_NARGS(func));
		longjmp(scm_errjmp, 1);
	  }
	  spush_opc(SCM_OP_CALLC0 + SCM_CPRIM_NARGS(func));
	  spush(func);
	  spush_opc(SCM_OP_PUSH);
	  needlink = FALSE;
	} else {					/* will take args directly from stack */
	  spush_opc(SCM_OP_CALLS);
	  spush(func);
	  spush_opc(SCM_OP_PUSH);
	}
	break;
  case SOBJ_T_PRIM:			/* primitive */
	spush(SCM_PRIM(func)->address);
	needlink = (SCM_PRIM(func)->nargs < 0);
	break;

  case SOBJ_T_SYNTAX:			/* syntax object. immediate execution */
	{
	  return((*SCM_CPRIM_FUNC(func))(arglist, env));
	}
  case SOBJ_T_MACRO:
	{
	  return(scm_syntax_object(scm_syntax_macro(func, form, env), env));
	}

  default:
	err("Don't know how to compile", func);
  }
  nargs = scm_syntax_args(arglist, env);

  if (needlink) {
	spush_opc(SCM_OP_MARK);
  }
  return(form);
}

SOBJ scm_syntax_object(SOBJ obj, SOBJ env)
{
  int depth;

  if (SCM_NNULLP(obj)) {
	switch(SCM_OBJTYPE(obj)) {
	case SOBJ_T_PAIR:	scm_syntax_pair(obj, env);				  	break;
	case SOBJ_T_INUM:
	case SOBJ_T_BNUM:
	case SOBJ_T_BOOLEAN:
	case SOBJ_T_FNUM: 	spush(obj);  spush_opc(SCM_OP_PUSH);	break;
	case SOBJ_T_ATOM:
	  {

		/* atom are generated by read and are converted to symbols
		 * according to current environement (environment also
		 * includes current module context) */
		
		SOBJ v;

		if (env) {				/* not in the global environement */
		  v = scm_env_search(env, obj, &depth);
		  if (v) {
			scm_syntax_lget(v, depth);
			break;
		  }
		}
		
		/* object must be found in the current module. */
		v = scm_module_find_symbol(scm_current_module, obj, FALSE);
		if (v) {
		  spush(v);	  spush_opc(SCM_OP_PUSHV);
		  break;
		}
		err("object not defined", obj);
	  }
	  
	case SOBJ_T_SYMBOL:
	  scm_puts("*** got symbol object in scm_syntax_object\n");
	  if (env) {
		SOBJ lvar = scm_env_search(env, SCM_SYM_NAME(obj), &depth);
		if (lvar) {				/* yeah... got a local variable */
		  scm_syntax_lget(lvar, depth);
		  break;
		}
	  }
	  spush(obj);
	  spush_opc(SCM_OP_PUSHV);
	  break;
	default:
	  spush(obj);  spush_opc(SCM_OP_PUSH);
	  break;
	  fprintf(stderr, "scm_syntax_object: unknow obj type %d\n", SCM_OBJTYPE(obj));
	}
  }
  return(obj);
}


/*-- compile on stack then create an code object and restore stack contents */

SOBJ scm_compile(SOBJ obj, SOBJ env)
{
  SOBJ code, *slimit;

  /*  scm_puts("; compile env:");	scm_cprint(scm_env); */
  
  slimit = scm_sp;
  spush_opc(SCM_OP_RETURN);
  scm_syntax_object(obj, env);
  /*  optimize_code(scm_sp, slimit); */
  code = scm_mkcode(scm_sp, slimit - scm_sp);
  scm_sp = slimit;
  return(code);
}
#endif /* OLD_COMPILER */
/********************************
 * executor
 ********************************/

/*-- err */
static char errbuf[256];

void err(char *msg, SOBJ obj)
{
#ifndef OLD
  PORT *p;
  SOBJ errstr;
  
  p = port_open_output_string();
  if (scm_inp) {
	port_puts(p, "\nerror: line ");	port_putn(p, scm_inp->line + 1);
	port_puts(p, ": ");
  }
  port_puts(p, msg);  port_puts(p, " ");  scm_write_obj(obj, p, TRUE);
  port_puts(p, "\n");

  if (scm_last_read_object != NULL) {
	port_puts(p, "  last read:");
	scm_write_obj(scm_last_read_object, p, TRUE);
	port_puts(p, "\n");
  }
  scm_last_read_object = NULL;
  errstr = scm_mkstring(port_string_output_string(p));
  port_close(p);
  scm_throw(scm_mkatom("error"), errstr);

#else
  if (scm_inp) {
	scm_puts("\nerror: line "); scm_putn(scm_inp->line + 1); scm_puts(": ");
  }
  scm_puts(msg);  scm_puts(" "); scm_cprint(obj);
  if (scm_last_read_object != NULL) {
	scm_puts("  last read:");  scm_write2(scm_last_read_object, NULL);
	scm_putc('\n');
  }
  scm_last_read_object = NULL;
  longjmp(scm_errjmp, 1);
#endif
}

SOBJ scm_error(SOBJ string, SOBJ object)
{
  if (!SCM_STRINGP(string)) err("error: bad string", string);
  err(SCM_STR_VALUE(string), object);
  return(NULL);					/* never returns */
}

SOBJ scm_execute(SOBJ code)
{
  SOBJ result;
  SOBJ *sp_save;

  if (!SCM_CODEP(code))	err("execute: bad code", code);
  sp_save = scm_sp;

  /*-- generate preamble code preamble on */
  spush_opc(SCM_OP_END);
  spush_opc(SCM_OP_CALL);
  spush(code);
  spush_opc(SCM_OP_PUSH);
  spush_opc(SCM_OP_MARK);

  spush(NULL);			/* marker: place holder for result */
  
  scm_start_engine();
  result = spop();
  scm_sp = sp_save;
  return(result);
}

SOBJ scm_eval(SOBJ expr, SOBJ env)
{
  return(scm_execute(scm_compile(expr, env)));
}


SOBJ scm_load(SOBJ filename)
{
  SOBJ obj, r;
  int k;
  jmp_buf   old_handler;
  PORT 		*old_input;

  old_input = NULL;

  if (!SCM_STRINGP(filename)) err("bad string", filename);

  /*-- save the old handler and install a new one */
  *old_handler = *scm_errjmp;
  old_input   = scm_inp;
  
  if ( (k = setjmp(scm_errjmp)) != 0) {
	if (scm_inp) port_close(scm_inp);	
	scm_inp = old_input;
	*scm_errjmp = *old_handler;
	err("load aborted", filename);
  }

  scm_inp = port_open_input_file(SCM_STR_VALUE(filename));
  if (scm_inp == NULL) {		/* not found */
	SOBJ tmpname;
	
	/* try if absolute path or if QS_LIB is not set */
	if (strchr(SCM_STR_VALUE(filename), '/') != NULL || scm_qs_lib == NULL) {
	  err("cannot open",
		  scm_string_append2(filename, scm_mkstring(", try to set QS_LIB")));
	}
	tmpname = scm_string_append2(scm_mkstring(scm_qs_lib),
								 scm_string_append2(scm_mkstring("/"),
													filename));
	scm_inp = port_open_input_file(SCM_STR_VALUE(tmpname));
  }

  if (scm_inp == NULL) 	err("cannot open", filename);
  r = NULL;
  while( (scm_last_read_object = obj = scm_read_port(scm_inp)) != scm_eof) {
	r = scm_execute(scm_compile(obj, NULL));
	/* scm_gc(); */
  }
  port_close(scm_inp);
  scm_inp  = old_input;
  *scm_errjmp  = *old_handler;
  return(r);
}

SOBJ scm_eval_string(SOBJ str)
{
  SOBJ obj, r;
  int k;
  jmp_buf   old_handler;
  PORT 		*old_input;

  old_input = NULL;

  if (!SCM_STRINGP(str)) err("bad string", str);

  /*-- save the old handler and install a new one */
  *old_handler = *scm_errjmp;
  old_input   = scm_inp;
  
  if ( (k = setjmp(scm_errjmp)) != 0) {
	if (scm_inp) port_close(scm_inp);	
	scm_inp = old_input;
	*scm_errjmp = *old_handler;
	err("load aborted", str);
  }

  scm_inp = port_open_input_string(SCM_STR_VALUE(str));
  if (scm_inp == NULL) 	err("cannot open", str);
  r = NULL;
  while( (scm_last_read_object = obj = scm_read_port(scm_inp)) != scm_eof) {
	r = scm_execute(scm_compile(obj, NULL));
	/* scm_gc(); */
  }
  port_close(scm_inp);
  scm_inp  = old_input;
  *scm_errjmp  = *old_handler;
  return(r);
}

/********************************
 * interpreter
 ********************************/
void scm_add_cprim(char *name, SCM_CPRIM f, int n)
{
  SOBJ sym, prim;

  prim = scm_newcell(SOBJ_T_CPRIM);
  SCM_CPRIM_FUNC(prim) = f;  SCM_CPRIM_NARGS(prim) = n;

  sym = scm_symadd(name, prim);
}

void scm_add_vmfunc(char *name, void (*f)(SCM_vmRegisters *))
{
  SOBJ sym, vmfunc;

  vmfunc = scm_newcell(SOBJ_T_VMFUNC);
  SCM_VMFUNC(vmfunc) = f;

  sym = scm_symadd(name, vmfunc);
}

void scm_add_prim(char *name, SCM_PRIM_TABLE *entry)
{
  SOBJ sym, prim;
  prim = scm_newcell(SOBJ_T_PRIM);
  SCM_PRIM(prim) = entry;

  sym = scm_symadd(name, prim);
}

SOBJ scm_add_syntax(char *name, SCM_CPRIM f)
{
  SOBJ sym, prim;
  prim = scm_newcell(SOBJ_T_SYNTAX);
  SCM_SYNTAX_FUNC(prim) = f;

  sym = scm_symadd(name, prim);  
  return(sym);
}

void scm_syntax_print(SOBJ obj, PORT *p)
{
  port_puts(p, "<#syntax>");
}

/*-- add a new type */

int scm_add_type(SOBJ_TYPE_DESCR *type)
{
  int type_nr = scm_type_next_descr++;
  scm_type_hook[type_nr] = *type;

  /*-- force no call for this type */
  scm_type_hook[type_nr].execute = (long)opcode(SCM_OP_NO_CALL);
  
  return(type_nr);
}


void scm_init_type()
{
  int i;
  for (i = 0; i < scm_type_next_descr; i++) {
	if (scm_type_hook[i].execute >= 0 &&
		scm_type_hook[i].execute < SCM_OP_MAX) {
	  scm_type_hook[i].execute = (long)(opcode(scm_type_hook[i].execute));
	}
  }
}


void scm_init()
{
  /* init heap and stack */
  scm_heap_init(scm_default_hs*1024);
  scm_stack_init(scm_default_ds*1024);

  /* capture important env variables */
  scm_qs_lib = getenv("QS_LIB");

  /* init the rest */
  scm_init_symbol_hash();
  scm_init_atom();				/* prepare to hold atoms */

  scm_true = scm_newcell(SOBJ_T_BOOLEAN);	   	scm_gc_protect(&scm_true);
  scm_false= scm_newcell(SOBJ_T_BOOLEAN);	   	scm_gc_protect(&scm_false);
  scm_unbound= scm_newcell(SOBJ_T_UNBOUND);	   	scm_gc_protect(&scm_unbound);
  scm_undefined= scm_newcell(SOBJ_T_UNDEFINED);	scm_gc_protect(&scm_undefined);

  scm_sym_unquote = scm_symadd("unquote", scm_unbound);
  scm_sym_unquote_splicing = scm_symadd("unquote-splicing", scm_unbound);

#ifdef OLD_COMPILER
  /*  scm_env = NULL;  scm_gc_protect(&scm_env); */

  /*-- syntax expressions */
  scm_sym_set 		= scm_add_syntax("set!", 		scm_syntax_set);
  scm_sym_quote		= scm_add_syntax("quote", 		scm_syntax_quote);
  scm_sym_qquote	= scm_add_syntax("quasiquote",	scm_syntax_qquote);
  scm_sym_if		= scm_add_syntax("if", 			scm_syntax_if);
  scm_sym_begin		= scm_add_syntax("begin", 		scm_syntax_begin);
  scm_sym_lambda	= scm_add_syntax("lambda", 		scm_syntax_lambda);
  scm_sym_code		= scm_add_syntax("code", 		scm_syntax_code);
  scm_sym_define	= scm_add_syntax("define", 		scm_syntax_define);
  scm_sym_let		= scm_add_syntax("let", 		scm_syntax_let);
  scm_sym_letstar	= scm_add_syntax("let*", 		scm_syntax_lets);
  scm_sym_letrec	= scm_add_syntax("letrec", 		scm_syntax_lets);
  scm_sym_env		= scm_add_syntax("the-env", 	scm_syntax_the_env);
  scm_sym_immediate = scm_add_syntax("immediate", 	scm_syntax_immediate);
  scm_sym_and		= scm_add_syntax("and",			scm_syntax_and);
  scm_sym_or		= scm_add_syntax("or",			scm_syntax_or);
  scm_sym_cond		= scm_add_syntax("cond",		scm_syntax_cond);
  scm_sym_catch		= scm_add_syntax("catch",		scm_syntax_catch);

  scm_sym_while		= scm_add_syntax("while",		NULL);

  scm_sym_else		= scm_symadd("else", scm_unbound);
#endif

  scm_add_cprim("disasm",		scm_disasm,			1);
  scm_add_cprim("read",			scm_read,			0);
  scm_add_cprim("error",		scm_error,			2);

  scm_init_port();
  scm_init_macro();
  scm_init_symbol();			/* now init symbol and keyword module */
  scm_init_boolean();
  scm_init_chr();
  scm_init_str();
  scm_init_list();
  scm_init_array();
  scm_init_hash();
  scm_init_variable();			/* generic variable system */
  scm_init_dyn();
  scm_init_catch();				/* catch/throw */

  scm_engine_init();			/* init the vm engine */
  scm_init_type();
  scm_init_module();			/* init the module system */
  scm_init_asm();				/* init the assembler */
  scm_init_misc();				/* init the misc system */

#ifdef OLD_COMPILER
  scm_cons_addr   = SCM_PRIM(SCM_SYM_VALUE(scm_symsearch("cons")))->address;
  scm_append_addr = SCM_PRIM(SCM_SYM_VALUE(scm_symsearch("append")))->address;
#endif

  if (!scm_no_init) {
	if (setjmp(scm_errjmp) == 0) {
	  scm_load(scm_mkstring("s.scm"));
	} else {
	  scm_puts("; load failed, trying to continue\n");
	}
  }
}

void scm_main_loop()
{
  SOBJ obj, code, result;
  SOBJ *oldsp;
  int k;

  k = setjmp(scm_errjmp);
  if (k != 0) {
	printf("toplevel restart: k=%d\n", k);
	scm_current_module = scm_global_module;
	fflush(stdout);
  }
  sclear();						/* top level: nothing on stacks */
  while(1) {
	if (sdepth() != 0) {
	  fprintf(stderr, "ERROR: Stack depth=%d. clearing stack\n", sdepth());
	  sclear();
	}
	scm_gc();
	scm_heap_stat();
	scm_putc('>');	scm_putc(' ');
	oldsp = scm_sp;

	scm_last_read_object = obj = scm_read_port(scm_inp);
	if (obj == scm_eof) 	break;
	code = scm_compile(obj, NULL);

#ifdef DEBUG_COMPILE
	scm_puts("*** stack after compile:");	sdump();
#endif

	result = scm_execute(code);
#ifdef DEBUG_REPL
	scm_puts("*** stack after execute:");	sdump();
#endif
	scm_write_obj(result, scm_outp, 0);		scm_putc('\n');
	scm_sp = oldsp;
#ifdef DEBUG_REPL
	scm_puts("*** stack after print:");		sdump();
#endif
  }
}

void scm_usage(char *prog)
{
  printf("Usage: %s [-h|--help] [-hs=#] [-ds=#] [-ni] ...\n", prog);
  printf("where:\n");
  printf("  -h | --help  print this help\n");
  printf("  -hs=#        set heap size to # * 1024 cells (default=%ld)\n",
		 scm_default_hs);
  printf("  -ds=#        set data stack to # * 1024 ptrs (default=%ld)\n",
		 scm_default_ds);
  printf("  -ni          don't load initialisation files\n");
  exit(1);
}	  

void scm_set_argv(int argc, char **argv)
{
  int n, i;
  SOBJ scm_argv;
  
  n = 0;
  for (i = 0 ; i < argc; i++) if (argv[i] != NULL) n++;

  scm_argv = scm_mkarray(n, NULL);
  n = 0;
  for (i = 0; i < argc; i++) {
	if (argv[i]) {
	  SCM_AREF(scm_argv, n) = scm_mkstring(argv[i]);
	  n++;
	}
  }
  scm_symadd("argv", scm_argv);
}

void scm_set_environ()
{
  extern char **environ;
  char **p, *q, *tmp;
  SOBJ envhash;

  envhash = scm_mkhash(SCM_HASH_T_GEN);
  p = environ;
  while(*p) {
	tmp = strdup(*p++);
	if ((q = strchr(tmp, '=')) == NULL) {
	  fprintf(stderr, "Bad environement entry: %s\n", tmp);
	} else {
	  *q++ = 0;
	  scm_hash_set(envhash, scm_mkstring(tmp), scm_mkstring(q));
	  /*  scm_puts("; hash after adding:"); scm_cprint(envhash); */
	}
	free(tmp);
  }
  scm_symadd("environ", envhash);
}


void scm_boot(int argc, char **argv)
{
  int stack_start;
  int i;

  scm_cstack_start = &stack_start;

  printf("; QScheme - A fast implementation of the Scheme language\n");
  printf("; Copyright (C) 1998-1999 Daniel Crettol <dan@sof.ch>\n");
  printf("; QScheme is distributed under the GNU General Public Licence.\n");
  printf("; See the LICENCE file for more informations.\n");
  for (i = 1; i < argc; i++) {
	if (strncmp(argv[i], "-hs=", 4) == 0) {
	  scm_default_hs = atol(argv[i]+4);
	  argv[i] = NULL;
	  continue;
	}
	if (strncmp(argv[i], "-ds=", 4) == 0) {
	  scm_default_ds = atol(argv[i]+4);
	  argv[i] = NULL;
	  continue;
	}
	if (strcmp(argv[i], "-ni") == 0) { 
	  scm_no_init=TRUE;
	  argv[i] = NULL;
	  continue; 
	}
	if (strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0) {
	  scm_usage(argv[0]);
	}
  }
  scm_port_init_default_files();
  scm_init();
  scm_set_argv(argc, argv);
  scm_set_environ();
  scm_main_loop();
}

